from typing import AsyncGenerator

from agentia.llm.completion import Listeners
from agentia.spec import *


class TextStream:
    def __init__(self, gen: AsyncGenerator[str, None]):
        async def _gen():
            async for part in gen:
                self.__all += part
                yield part

        self.__all = ""
        self.__gen = _gen()

    def __aiter__(self) -> AsyncGenerator[str, None]:
        return self.__gen

    async def __wait_for_completion(self) -> str:
        async for _ in self.__gen:
            pass
        return self.__all

    def __await__(self):
        return self.__wait_for_completion().__await__()


class ReasoningStream(TextStream):
    type: Literal["reasoning-stream"] = "reasoning-stream"


class MessageStream(TextStream):
    type: Literal["message-stream"] = "message-stream"


class ChatCompletionStreamBase:
    def __init__(self) -> None:
        self.usage = Usage()
        self.finish_reason: FinishReason | None = None
        self.warnings: list[Warning] = []
        self.new_messages: list[Message] = []
        self.on_finish = Listeners()
        self.on_new_message = Listeners()

    def add_new_message(self, msg: Message):
        self.new_messages.append(msg)
        self.on_new_message.emit(msg)


class ChatCompletionStream(ChatCompletionStreamBase):
    def __init__(self, gen: AsyncGenerator[StreamPart, None]):
        super().__init__()

        async def __gen() -> AsyncGenerator[
            ReasoningStream | MessageStream | Source | File | ToolCall | ToolResult,
            None,
        ]:
            async for part in gen:
                if isinstance(
                    part,
                    (
                        SourceURL,
                        SourceDocument,
                        File,
                        ToolCall,
                        ToolResult,
                    ),
                ):
                    yield part
                elif isinstance(part, StreamPartTextStart):

                    async def gen2():
                        async for part2 in gen:
                            if isinstance(part2, StreamPartTextDelta):
                                yield part2.delta
                            elif isinstance(part2, StreamPartTextEnd):
                                break

                    yield MessageStream(gen2())
                elif isinstance(part, StreamPartReasoningStart):

                    async def gen2():
                        async for part2 in gen:
                            if isinstance(part2, StreamPartReasoningDelta):
                                yield part2.delta
                            elif isinstance(part2, StreamPartReasoningEnd):
                                break

                    yield ReasoningStream(gen2())
            self.on_finish.emit()

        self.__gen = __gen()

    def __aiter__(
        self,
    ) -> AsyncGenerator[
        ReasoningStream | MessageStream | Source | File | ToolCall | ToolResult, None
    ]:
        return self.__gen

    async def __wait_for_completion(self) -> AssistantMessage:
        async for item in self.__gen:
            ...
        m = self.new_messages[-1]
        assert isinstance(m, AssistantMessage)
        return m

    def __await__(self):
        return self.__wait_for_completion().__await__()


class ChatCompletionEvents(ChatCompletionStreamBase):
    def __init__(self, gen: AsyncGenerator[StreamPart, None]):
        super().__init__()

        async def __gen():
            async for part in gen:
                yield part
            self.on_finish.emit()

        self.__gen = __gen()

    def __aiter__(self) -> AsyncGenerator[StreamPart, None]:
        return self.__gen

    async def __wait_for_completion(self) -> AssistantMessage:
        async for item in self.__gen:
            ...
        m = self.new_messages[-1]
        assert isinstance(m, AssistantMessage)
        return m

    def __await__(self):
        return self.__wait_for_completion().__await__()



import os
from typing import Callable, Type
from ..utils.decorators import tool
from agentia.tools import Plugin, PluginInitError


ALL_PLUGINS: dict[str, type[Plugin]] = {}


__all__ = [
    "Plugin",
    "PluginInitError",
    "ALL_PLUGINS",
    "tool",
]

if os.environ.get("AGENTIA_DISABLE_PLUGINS", "").lower() not in [
    "1",
    "true",
    "yes",
    "y",
]:
    try:
        from .calc import CalculatorPlugin
        from .clock import ClockPlugin
        from .code import CodePlugin
        from .memory import MemoryPlugin
        from .search import SearchPlugin
        from .web import WebPlugin

        ALL_PLUGINS = {
            "calc": CalculatorPlugin,
            "clock": ClockPlugin,
            "code": CodePlugin,
            "memory": MemoryPlugin,
            "search": SearchPlugin,
            "web": WebPlugin,
        }
    except ImportError as e:
        # print("Failed to import built-in plugins:", e)
        pass
        # raise e


for name, cls in ALL_PLUGINS.items():
    cls._BUILTIN_ID = name


def register_plugin(
    name: str, overwrite: bool = False
) -> Callable[[Type[Plugin]], Type[Plugin]]:
    if not overwrite:
        assert name not in ALL_PLUGINS, f"Plugin {name} already registered"

    def wrapper(cls: Type[Plugin]):
        ALL_PLUGINS[name] = cls
        cls._BUILTIN_ID = name
        return cls

    return wrapper



import os
from pathlib import Path
from typing import Literal, Sequence, Union, overload

import uuid

from agentia.history import History
from agentia.llm import LLM, GenerationOptions
from agentia.llm.completion import ChatCompletion
from agentia.llm.stream import ChatCompletionStream, ChatCompletionEvents
from agentia.spec.chat import AssistantMessage, ToolMessage
from agentia.tools.tools import Tool, ToolSet
from agentia.spec import (
    NonSystemMessage,
    UserMessage,
    MessagePartText,
    ObjectType,
)
import logging


class Agent:
    def __init__(
        self,
        model: str | None = None,
        tools: Sequence[Tool] | None = None,
        id: str | None = None,
        name: str | None = None,
        icon: str | None = None,
        description: str | None = None,
        instructions: str | None = None,
        options: GenerationOptions | None = None,
    ) -> None:
        self.id = str(id or uuid.uuid4())
        self.name = name
        self.icon = icon
        self.description = description
        options = options or GenerationOptions()
        if tools:
            options["tools"] = ToolSet(tools)
        self.options = options
        if not model:
            model = os.getenv("AGENTIA_DEFAULT_MODEL", "openai/gpt-5-mini")
        self.llm = LLM(model)
        self.llm._agent = self
        self.history = History()
        if instructions:
            self.history.add_instructions(instructions)
        self.log = logging.getLogger(f"agentia.agent")

    @staticmethod
    def from_config(config: Union[str, Path, "Config"]) -> "Agent":
        from agentia.utils.config import load_agent_from_config

        return load_agent_from_config(config)

    def __add_prompt(
        self,
        prompt: str | NonSystemMessage | Sequence[NonSystemMessage],
    ) -> None:
        if isinstance(prompt, str):
            self.history.add(UserMessage(content=[MessagePartText(text=prompt)]))
        elif not isinstance(prompt, (list, Sequence)):
            self.history.add(prompt)
        else:
            self.history.add(*prompt)

    @overload
    def run(
        self,
        prompt: str | NonSystemMessage | Sequence[NonSystemMessage],
        /,
        stream: Literal[False] = False,
        events: Literal[False] = False,
    ) -> ChatCompletion: ...

    @overload
    def run(
        self,
        prompt: str | NonSystemMessage | Sequence[NonSystemMessage],
        /,
        stream: Literal[True],
        events: Literal[False] = False,
    ) -> ChatCompletionStream: ...

    @overload
    def run(
        self,
        prompt: str | NonSystemMessage | Sequence[NonSystemMessage],
        /,
        stream: Literal[True],
        events: Literal[True],
    ) -> ChatCompletionEvents: ...

    def run(
        self,
        prompt: str | NonSystemMessage | Sequence[NonSystemMessage],
        /,
        stream: bool = False,
        events: bool = False,
    ) -> ChatCompletion | ChatCompletionStream | ChatCompletionEvents:
        self.__add_prompt(prompt)
        if stream:
            x = self.llm.stream(self.history.get(), events=events, options=self.options)
        else:
            assert not events, "events=True is only supported with stream=True"
            x = self.llm.generate(self.history.get(), options=self.options)

        def on_finish():
            for m in x.new_messages:
                assert isinstance(m, (AssistantMessage, ToolMessage, UserMessage))
                self.history.add(m)

        x.on_finish.on(on_finish)

        return x

    async def generate_object[T: ObjectType](
        self,
        prompt: str | NonSystemMessage | Sequence[NonSystemMessage],
        type: type[T],
    ) -> T:
        self.__add_prompt(prompt)
        msgs = await self.llm._generate_object_impl(
            self.history.get(), type, options=self.options
        )

        for m in msgs:
            assert isinstance(m, (AssistantMessage, ToolMessage, UserMessage))
            self.history.add(m)

        assert isinstance(msgs[-1], AssistantMessage)

        return msgs[-1].parse(type)


from .utils.config import Config



from __future__ import annotations

import logging

LOGGER = logging.getLogger("agentia")


def init_logging(level: logging._Level = logging.INFO):
    logging.basicConfig(
        level=level, format="[%(asctime)s][%(levelname)-8s][%(name)s] %(message)s"
    )
    logging.getLogger("httpx").setLevel(logging.WARNING)


from dotenv import load_dotenv

from . import utils

from .agent import Agent
from .spec import (
    Message,
    UserMessage,
    AssistantMessage,
    SystemMessage,
    MessagePart,
    MessagePartText,
    MessagePartFile,
    MessagePartReasoning,
    MessagePartToolCall,
    MessagePartToolResult,
    StreamPart,
    ToolCall,
    ToolResult,
)
from .utils.decorators import magic, ImageUrl
from .utils.decorators import tool
from .plugins import Plugin, PluginInitError, register_plugin, ALL_PLUGINS
from . import llm, plugins
from .tools import ProviderTool, Tool, Tools, MCPServer, MCPContext, mcp_context
from .history import History
from .llm.completion import ChatCompletion
from .llm.stream import (
    TextStream,
    ReasoningStream,
    MessageStream,
    ChatCompletionStream,
    ChatCompletionEvents,
)

__all__ = [
    # submodules
    "plugins",
    "utils",
    "llm",
    # agent
    "Agent",
    "History",
    # llm
    "ChatCompletion",
    "TextStream",
    "ReasoningStream",
    "MessageStream",
    "ChatCompletionStream",
    "ChatCompletionEvents",
    # spec
    "Message",
    "SystemMessage",
    "UserMessage",
    "AssistantMessage",
    "MessagePart",
    "MessagePartText",
    "MessagePartFile",
    "MessagePartReasoning",
    "MessagePartToolCall",
    "MessagePartToolResult",
    "StreamPart",
    "ToolCall",
    "ToolResult",
    # plugins
    "Plugin",
    "PluginInitError",
    "ToolResult",
    "register_plugin",
    # utils
    "init_logging",
    "magic",
    "tool",
    "ImageUrl",
    "ProviderTool",
    "Tool",
    "Tools",
    "MCPServer",
    "MCPContext",
    "mcp_context",
    "load_dotenv",
]



from ._openai_api import OpenAIAPIProvider
import os


class Gateway(OpenAIAPIProvider):
    def __init__(
        self, model: str, api_key: str | None = None, base_url: str | None = None
    ):
        api_key = api_key or os.environ.get("AI_GATEWAY_API_KEY")
        if not api_key:
            raise ValueError("AI_GATEWAY_API_KEY environment variable not set")
        base_url = "https://ai-gateway.vercel.sh/v1"
        super().__init__(
            provider="gateway", model=model, api_key=api_key, base_url=base_url
        )
        self.extra_body["modalities"] = ["text", "image"]



import os
from . import Plugin, tool
from typing import Annotated
import httpx
from markdownify import markdownify
from tavily import AsyncTavilyClient


class WebPlugin(Plugin):
    def __init__(self, tavily_api_key: str | None = None):
        api_key = tavily_api_key or os.environ.get("TAVILY_API_KEY")
        if not api_key:
            raise ValueError("TAVILY_API_KEY is required for WebPlugin")
        self.__tavily = AsyncTavilyClient(api_key=api_key)

    async def __get(self, url: str):
        async with httpx.AsyncClient(
            headers={
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
            }
        ) as client:
            res = await client.get(url)
            res.raise_for_status()
            content_type = res.headers.get("content-type")
            if content_type == "text/html":
                md = markdownify(res.text)
                return {"content": md, "content_type": content_type}
            # raw text: .txt, .csv, .json, .md
            if content_type in ["text/plain", "application/json", "text/markdown"]:
                return {"content": res.text, "content_type": content_type}
            # If the content is not supported, return the raw content
            return {"content": res.text, "content_type": content_type}

    @tool
    async def get_webpage_content(
        self,
        url: Annotated[str, "The URL of the web page to get the content of"],
    ):
        """
        Access a web page by a URL, and fetch the content of this web page (in markdown format).
        You can always use this tool to directly access web content or access external sites.
        Use it at any time when you think you may need to access the internet.
        """
        result = await self.__tavily.extract(
            urls=url,
            # extract_depth="advanced",
            include_images=True,
        )
        failed_results = result.get("failed_results", [])
        if len(failed_results) > 0:
            try:
                return await self.__get(url)
            except Exception as e:
                pass
        return result



from contextlib import AsyncExitStack
from pathlib import Path
from typing import Any, Callable, Literal, overload
from typing import TYPE_CHECKING, Sequence
from mcp import ClientSession, StdioServerParameters, stdio_client, Tool as MCPTool
from mcp.client.websocket import websocket_client
from mcp.client.streamable_http import streamablehttp_client
from mcp.client.sse import sse_client
from pydantic import BaseModel
import asyncio
from httpx import Auth


if TYPE_CHECKING:  # pragma: no cover
    from agentia.tools.tools import _MCPTool


class MCPServerConfig(BaseModel):
    command: str
    args: list[str]
    env: dict[str, str] | None = None


class MCPContext:
    def __init__(self):
        self.exit_stack = AsyncExitStack()
        self.servers: list[MCPServer] = []

    async def __aenter__(self):
        global MCP_CONTEXTS
        MCP_CONTEXTS.append(self)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        global MCP_CONTEXTS
        MCP_CONTEXTS.remove(self)
        for server in self.servers:
            server.context_available = False
        await self.exit_stack.aclose()


MCP_CONTEXTS: list[MCPContext] = []


def mcp_context(f: Callable[..., Any]) -> Callable[..., Any]:
    """
    Context manager for MCP server initialization.
    This ensures that the MCP server is initialized and cleaned up properly.
    """
    assert asyncio.iscoroutinefunction(f), "Function must be async"

    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        async with MCPContext() as ctx:
            return await f(*args, **kwargs)

    return wrapper


def _convert_tool_format(tool: MCPTool) -> Any:
    converted_tool = {
        "name": tool.name,
        "description": tool.description,
        "parameters": {
            "type": "object",
            "properties": tool.inputSchema["properties"],
            "required": tool.inputSchema.get("required", []),
        },
    }
    return converted_tool


class MCPServer:
    @overload
    def __init__(
        self,
        name: str,
        *,
        command: str,
        args: Sequence[str | Path],
        env: dict[str, str] | None = None,
        cwd: str | Path | None = None,
        type: Literal["local"] = "local",
    ): ...

    @overload
    def __init__(
        self,
        name: str,
        *,
        type: Literal["http", "sse"],
        url: str,
        headers: dict[str, str] | None = None,
        auth: Auth | None = None,
    ): ...

    @overload
    def __init__(
        self,
        name: str,
        *,
        type: Literal["websocket"],
        url: str,
    ): ...

    def __init__(
        self,
        name: str,
        *,
        command: str | None = None,
        args: Sequence[str | Path] | None = None,
        env: dict[str, str] | None = None,
        cwd: str | Path | None = None,
        url: str | None = None,
        headers: dict[str, str] | None = None,
        auth: Auth | None = None,
        type: Literal["websocket", "sse", "http", "local"] = "local",
    ):

        self.cmd = None
        if command:
            self.cmd = [command] + ([a for a in args] if args else [])
        self.env = env
        self.cwd = cwd
        self.url = url
        self.headers = headers
        self.auth = auth
        self.type = type

        self.name = name
        self.initialized = False
        self.context_available = False
        self.__tools: list["_MCPTool"] = []
        self.session: ClientSession | None = None

        def assert_no_local_args():
            assert command is None, "Remote MCP server does not support command"
            assert args is None, "Remote MCP server does not support args"
            assert env is None, "Remote MCP server does not support env"
            assert cwd is None, "Remote MCP server does not support cwd"

        match type:
            case "local":
                assert url is None, "Local MCP server does not support URL"
                assert headers is None, "Local MCP server does not support headers"
                assert auth is None, "Local MCP server does not support auth"
                assert (
                    command is not None and len(command) > 0
                ), "Command must be specified"
                assert args is not None, "Args must be specified"
            case "http" | "sse":
                assert_no_local_args()
                assert url is not None, "URL must be specified for HTTP server"
            case "websocket":
                assert_no_local_args()
                assert url is not None, "URL must be specified for WebSocket server"
                assert headers is None, "WebSocket does not support headers"
                assert auth is None, "WebSocket does not support auth"

    def get_tools(self) -> list["_MCPTool"]:
        assert self.initialized, "MCP Server not initialized"
        return self.__tools

    async def init(self):
        from agentia.tools.tools import _MCPTool

        assert not self.initialized, "MCP Server already initialized"
        assert len(MCP_CONTEXTS) > 0, "Agents must be running in an MCP context"

        exit_stack = MCP_CONTEXTS[-1].exit_stack
        self.context_available = True
        MCP_CONTEXTS[-1].servers.append(self)

        match self.type:
            case "local":
                assert self.cmd is not None
                server_params = StdioServerParameters(
                    command=str(self.cmd[0]),
                    args=[str(a) for a in self.cmd[1:]],
                    env=self.env,
                    cwd=self.cwd,
                )
                self.stdio, self.write = await exit_stack.enter_async_context(
                    stdio_client(server_params)
                )
            case "websocket":
                assert self.url is not None
                self.stdio, self.write = await exit_stack.enter_async_context(
                    websocket_client(self.url)
                )
            case "sse":
                assert self.url is not None
                self.stdio, self.write = await exit_stack.enter_async_context(
                    sse_client(
                        self.url,
                        headers=self.headers,
                        auth=self.auth,
                    )
                )
            case "http":
                assert self.url is not None
                self.stdio, self.write, _ = await exit_stack.enter_async_context(
                    streamablehttp_client(
                        self.url, headers=self.headers, auth=self.auth
                    )
                )
            case _:
                raise ValueError(f"Unsupported MCP server type: {self.type}")
        self.session = await exit_stack.enter_async_context(
            ClientSession(self.stdio, self.write)
        )
        await self.session.initialize()
        tools = await self.session.list_tools()
        for tool in tools.tools:
            schema = _convert_tool_format(tool)
            self.__tools.append(_MCPTool(name=tool.name, schema=schema, server=self))
        while tools.nextCursor:
            tools = await self.session.list_tools(cursor=tools.nextCursor)
            for tool in tools.tools:
                schema = _convert_tool_format(tool)
                self.__tools.append(
                    _MCPTool(name=tool.name, schema=schema, server=self)
                )
        self.initialized = True

    async def run(self, tool: str, args: Any) -> Any:
        assert self.context_available, "Invalid or closed MCP context"
        assert self.session is not None, "MCP session not initialized"
        result = await self.session.call_tool(tool, args)
        assert self.context_available, "Invalid or closed MCP context"
        result_json = []
        for part in result.content:
            result_json.append(part.model_dump())
        return result_json



import abc
import re
from dataclasses import dataclass
from typing import TYPE_CHECKING, AsyncGenerator, Sequence
from agentia.llm import GenerationOptions
from agentia.tools.tools import ToolSet
import agentia.spec as spec

if TYPE_CHECKING:
    from .. import LLM


@dataclass
class ProviderGenerationResult:
    message: spec.AssistantMessage
    tool_calls: list[spec.ToolCall]
    finish_reason: spec.FinishReason
    usage: spec.Usage
    warnings: Sequence[spec.Warning]
    provider_metadata: spec.ProviderMetadata | None


class Provider(abc.ABC):
    provider: str
    model: str
    supported_urls: dict[str, Sequence[re.Pattern | str]]
    """
    Supported URL patterns by media type for the provider.

    The keys are media type patterns or full media types (e.g. `*\\/*` for everything, `audio/*`, `video/*`, or `application/pdf`).
    and the values are arrays of regular expressions that match the URL paths.

    The matching should be against lower-case URLs.

    Matched URLs are supported natively by the model and are not downloaded.
    """
    llm: "LLM"

    def __init__(self, provider: str, model: str):
        self.provider = provider
        self.model = model
        self.supported_urls = {}

    @abc.abstractmethod
    async def do_generate(
        self, prompt: spec.Prompt, tool_set: ToolSet, options: GenerationOptions
    ) -> ProviderGenerationResult: ...

    @abc.abstractmethod
    def do_stream(
        self, prompt: spec.Prompt, tool_set: ToolSet, options: GenerationOptions
    ) -> AsyncGenerator[spec.StreamPart, None]: ...



from datetime import datetime
from typing import Annotated, Any, Literal, Sequence
from pydantic import AliasChoices, BaseModel, Field, JsonValue, HttpUrl

type ProviderOptions = dict[str, dict[str, JsonValue]]
type ProviderMetadata = dict[str, dict[str, JsonValue]]

type DataContent = bytes | str | HttpUrl

type FinishReason = Literal[
    "stop", "length", "content-filter", "tool-calls", "error", "other", "unknown"
]


class ToolChoiceAuto(BaseModel):
    type: Literal["auto"] = "auto"


class ToolChoiceNone(BaseModel):
    type: Literal["none"] = "none"


class ToolChoiceRequired(BaseModel):
    type: Literal["required"] = "required"


class ToolChoiceSpecific(BaseModel):
    type: Literal["tool"] = "tool"
    tool_name: str = Field(
        validation_alias=AliasChoices("toolName", "tool_name"),
        serialization_alias="toolName",
    )


type ToolChoice = Annotated[
    ToolChoiceAuto | ToolChoiceNone | ToolChoiceRequired | ToolChoiceSpecific,
    Field(discriminator="type"),
]


class Usage(BaseModel):
    input_tokens: int | None = Field(
        default=None,
        validation_alias=AliasChoices("inputTokens", "input_tokens"),
        serialization_alias="inputTokens",
    )
    output_tokens: int | None = Field(
        default=None,
        validation_alias=AliasChoices("outputTokens", "output_tokens"),
        serialization_alias="outputTokens",
    )
    total_tokens: int | None = Field(
        default=None,
        validation_alias=AliasChoices("totalTokens", "total_tokens"),
        serialization_alias="totalTokens",
    )
    reasoning_tokens: int | None = Field(
        default=None,
        validation_alias=AliasChoices("reasoningTokens", "reasoning_tokens"),
        serialization_alias="reasoningTokens",
    )
    cached_input_tokens: int | None = Field(
        default=None,
        validation_alias=AliasChoices("cachedInputTokens", "cached_input_tokens"),
        serialization_alias="cachedInputTokens",
    )

    def __add__(self, y: "Usage") -> "Usage":
        x = Usage(**self.model_dump())
        if y.input_tokens:
            x.input_tokens = (x.input_tokens or 0) + y.input_tokens
        if y.output_tokens:
            x.output_tokens = (x.output_tokens or 0) + y.output_tokens
        if y.total_tokens:
            x.total_tokens = (x.total_tokens or 0) + y.total_tokens
        if y.reasoning_tokens:
            x.reasoning_tokens = (x.reasoning_tokens or 0) + y.reasoning_tokens
        if y.cached_input_tokens:
            x.cached_input_tokens = (x.cached_input_tokens or 0) + y.cached_input_tokens
        return x


class ResponseMetadata(BaseModel):
    id: str | None = None
    timestamp: datetime | None = None
    model_id: str | None = Field(
        default=None,
        validation_alias=AliasChoices("modelId", "model_id"),
        serialization_alias="modelId",
    )


class FunctionTool(BaseModel):
    type: Literal["function"] = "function"
    name: str
    input_schema: JsonValue = Field(
        validation_alias=AliasChoices("inputSchema", "input_schema"),
        serialization_alias="inputSchema",
    )
    description: str | None = None
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "provider_options"),
        serialization_alias="providerOptions",
    )

    def to_openai_schema(self) -> Any:
        return {
            "type": self.type,
            "name": self.name,
            "description": self.description,
            "parameters": self.input_schema,
            "strict": True,
        }


class ProviderDefinedTool(BaseModel):
    type: Literal["provider-defined"] = "provider-defined"
    id: str
    name: str
    args: dict[str, JsonValue]

    def to_openai_schema(self) -> Any:
        return {"type": self.id, **self.args}


type Tool = Annotated[FunctionTool | ProviderDefinedTool, Field(discriminator="type")]


class UnsupportedSettingWarning(BaseModel):
    type: Literal["unsupported-setting"] = "unsupported-setting"
    setting: str
    details: str | None = None


class UnsupportedToolWarning(BaseModel):
    type: Literal["unsupported-tool"] = "unsupported-tool"
    tool: Tool
    details: str | None = None


class OtherWarning(BaseModel):
    type: Literal["other"] = "other"
    message: str


type Warning = Annotated[
    UnsupportedSettingWarning | UnsupportedToolWarning | OtherWarning,
    Field(discriminator="type"),
]


class ToolCall(BaseModel):
    type: Literal["tool-call"] = "tool-call"

    tool_call_id: str = Field(
        validation_alias=AliasChoices("toolCallId", "tool_call_id"),
        serialization_alias="toolCallId",
    )
    """The identifier of the tool call. It must be unique across all tool calls."""

    tool_name: str = Field(
        validation_alias=AliasChoices("toolName", "tool_name"),
        serialization_alias="toolName",
    )
    """The name of the tool that should be called."""

    input: str
    """Stringified JSON object with the tool call arguments. Must match the parameters schema of the tool."""

    provider_executed: bool | None = Field(
        default=None,
        validation_alias=AliasChoices("providerExecuted", "provider_executed"),
        serialization_alias="providerExecuted",
    )
    """
    Whether the tool call will be executed by the provider.
    If this flag is not set or is false, the tool call will be executed by the client.
    """

    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )
    """Additional provider-specific metadata for the tool call."""


class ToolResult(BaseModel):
    type: Literal["tool-result"] = "tool-result"

    tool_call_id: str = Field(
        validation_alias=AliasChoices("toolCallId", "tool_call_id"),
        serialization_alias="toolCallId",
    )
    """The ID of the tool call that this result is associated with."""

    tool_name: str = Field(
        validation_alias=AliasChoices("toolName", "tool_name"),
        serialization_alias="toolName",
    )
    """Name of the tool that generated this result."""

    result: JsonValue
    """Result of the tool call. This is a JSON-serializable object."""

    is_error: bool | None = Field(
        default=None,
        validation_alias=AliasChoices("isError", "is_error"),
        serialization_alias="isError",
    )
    """Optional flag if the result is an error or an error message."""

    provider_executed: bool | None = Field(
        default=None,
        validation_alias=AliasChoices("providerExecuted", "provider_executed"),
        serialization_alias="providerExecuted",
    )
    """
    Whether the tool result was generated by the provider.
    If this flag is set to true, the tool result was generated by the provider.
    If this flag is not set or is false, the tool result was generated by the client.
    """

    preliminary: bool | None = None
    """
    Whether the tool result is preliminary.

    Preliminary tool results replace each other, e.g. image previews.
    There always has to be a final, non-preliminary tool result.

    If this flag is set to true, the tool result is preliminary.
    If this flag is not set or is false, the tool result is not preliminary.
    """

    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )
    """Additional provider-specific metadata for the tool result."""


class File(BaseModel):
    type: Literal["file"] = "file"

    media_type: str = Field(
        validation_alias=AliasChoices("mediaType", "media_type"),
        serialization_alias="mediaType",
    )
    """
    The IANA media type of the file, e.g. `image/png` or `audio/mp3`.

    @see https://www.iana.org/assignments/media-types/media-types.xhtml
    """

    data: str | bytes
    """
    Generated file data as base64 encoded strings or binary data.

    The file data should be returned without any unnecessary conversion.
    If the API returns base64 encoded strings, the file data should be returned
    as base64 encoded strings. If the API returns binary data, the file data should
    be returned as binary data.
    """


class SourceURL(BaseModel):
    type: Literal["source"] = "source"
    source_type: Literal["url"] = Field(
        "url",
        validation_alias=AliasChoices("sourceType", "source_type"),
        serialization_alias="sourceType",
    )
    id: str
    url: str
    title: str | None = None
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class SourceDocument(BaseModel):
    type: Literal["source"] = "source"
    source_type: Literal["document"] = Field(
        "document",
        validation_alias=AliasChoices("sourceType", "source_type"),
        serialization_alias="sourceType",
    )
    id: str
    media_type: str = Field(
        validation_alias=AliasChoices("mediaType", "media_type"),
        serialization_alias="mediaType",
    )
    title: str
    filename: str | None = None
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


type Source = Annotated[SourceURL | SourceDocument, Field(discriminator="source_type")]


class Reasoning(BaseModel):
    type: Literal["reasoning"] = "reasoning"
    text: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class Text(BaseModel):
    type: Literal["text"] = "text"
    text: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


type Content = Annotated[
    Text | Reasoning | File | Source | ToolCall | ToolResult,
    Field(discriminator="type"),
]



from ._openai_api import OpenAIAPIProvider
import os

# Model list: https://www.alibabacloud.com/help/en/model-studio/getting-started/models


class Chutes(OpenAIAPIProvider):
    def __init__(
        self, model: str, api_key: str | None = None, base_url: str | None = None
    ):
        api_key = (
            api_key
            or os.environ.get("CHUTES_API_KEY")
            or os.environ.get("CHUTES_API_TOKEN")
        )
        if not api_key:
            raise ValueError(
                "CHUTES_API_KEY or CHUTES_API_TOKEN environment variable not set"
            )
        base_url = "https://llm.chutes.ai/v1"
        super().__init__(
            provider="chutes", model=model, api_key=api_key, base_url=base_url
        )



from typing import AsyncGenerator

from agentia.spec.base import FinishReason, Usage, Warning
from agentia.spec.chat import AssistantMessage, Message, ToolMessage


class Listeners:
    def __init__(self):
        self.__listeners = []

    def on(self, listener):
        self.__listeners.append(listener)

    def off(self, listener):
        self.__listeners.remove(listener)

    def emit(self, *args, **kwargs):
        for listener in self.__listeners:
            listener(*args, **kwargs)


class ChatCompletion:
    def __init__(
        self,
        gen: AsyncGenerator[AssistantMessage | ToolMessage, None],
    ):
        async def __gen() -> AsyncGenerator[AssistantMessage | ToolMessage, None]:
            async for msg in gen:
                yield msg
            self.on_finish.emit()

        self.__gen = __gen()
        self.usage = Usage()
        self.finish_reason: FinishReason | None = None
        self.warnings: list[Warning] = []
        self.new_messages: list[Message] = []
        self.on_finish = Listeners()
        self.on_new_message = Listeners()

    def add_new_message(self, msg: Message):
        self.new_messages.append(msg)
        self.on_new_message.emit(msg)

    def __aiter__(self) -> AsyncGenerator[AssistantMessage | ToolMessage, None]:
        return self.__gen

    async def __wait_for_completion(self) -> AssistantMessage:
        async for msg in self.__gen:
            ...
        m = self.new_messages[-1]
        assert isinstance(m, AssistantMessage)
        return m

    def __await__(self):
        return self.__wait_for_completion().__await__()
